import { AggregateRoot, AggregateID } from '@mobicoop/ddd-library';
import { v4 } from 'uuid';
import { AdCreatedDomainEvent } from './events/ad-created.domain-events';
import { AdProps, CreateAdProps, DefaultAdProps } from './ad.types';
import { Waypoint } from './value-objects/waypoint.value-object';
import { MarginDurationsProps } from './value-objects/margin-durations.value-object';

export class AdEntity extends AggregateRoot<AdProps> {
  protected readonly _id: AggregateID;

  static create = (
    create: CreateAdProps,
    defaultAdProps: DefaultAdProps,
  ): AdEntity => {
    const id = v4();
    const props: AdProps = { ...create };
    const ad = new AdEntity({ id, props })
      .setMissingMarginDurations(defaultAdProps.marginDurations)
      .setMissingStrict(defaultAdProps.strict)
      .setDefaultDriverAndPassengerParameters({
        driver: defaultAdProps.driver,
        passenger: defaultAdProps.passenger,
        seatsProposed: defaultAdProps.seatsProposed,
        seatsRequested: defaultAdProps.seatsRequested,
      })
      .setMissingWaypointsPosition();
    ad.addEvent(
      new AdCreatedDomainEvent({
        aggregateId: id,
        userId: props.userId,
        driver: props.driver,
        passenger: props.passenger,
        frequency: props.frequency,
        fromDate: props.fromDate,
        toDate: props.toDate,
        monTime: props.schedule.mon,
        tueTime: props.schedule.tue,
        wedTime: props.schedule.wed,
        thuTime: props.schedule.thu,
        friTime: props.schedule.fri,
        satTime: props.schedule.sat,
        sunTime: props.schedule.sun,
        monMarginDuration: props.marginDurations.mon,
        tueMarginDuration: props.marginDurations.tue,
        wedMarginDuration: props.marginDurations.wed,
        thuMarginDuration: props.marginDurations.thu,
        friMarginDuration: props.marginDurations.fri,
        satMarginDuration: props.marginDurations.sat,
        sunMarginDuration: props.marginDurations.sun,
        seatsProposed: props.seatsProposed,
        seatsRequested: props.seatsRequested,
        strict: props.strict,
        waypoints: props.waypoints.map((waypoint: Waypoint) => ({
          position: waypoint.position,
          name: waypoint.address.name,
          houseNumber: waypoint.address.houseNumber,
          street: waypoint.address.street,
          postalCode: waypoint.address.postalCode,
          locality: waypoint.address.locality,
          country: waypoint.address.postalCode,
          lon: waypoint.address.coordinates.lon,
          lat: waypoint.address.coordinates.lat,
        })),
      }),
    );
    return ad;
  };

  private setMissingMarginDurations = (
    defaultMarginDurations: MarginDurationsProps,
  ): AdEntity => {
    if (!this.props.marginDurations) this.props.marginDurations = {};
    if (!this.props.marginDurations.mon)
      this.props.marginDurations.mon = defaultMarginDurations.mon;
    if (!this.props.marginDurations.tue)
      this.props.marginDurations.tue = defaultMarginDurations.tue;
    if (!this.props.marginDurations.wed)
      this.props.marginDurations.wed = defaultMarginDurations.wed;
    if (!this.props.marginDurations.thu)
      this.props.marginDurations.thu = defaultMarginDurations.thu;
    if (!this.props.marginDurations.fri)
      this.props.marginDurations.fri = defaultMarginDurations.fri;
    if (!this.props.marginDurations.sat)
      this.props.marginDurations.sat = defaultMarginDurations.sat;
    if (!this.props.marginDurations.sun)
      this.props.marginDurations.sun = defaultMarginDurations.sun;
    return this;
  };

  private setMissingStrict = (strict: boolean): AdEntity => {
    if (this.props.strict === undefined) this.props.strict = strict;
    return this;
  };

  private setDefaultDriverAndPassengerParameters = (
    defaultDriverAndPassengerParameters: DefaultDriverAndPassengerParameters,
  ): AdEntity => {
    this.props.driver = !!this.props.driver;
    this.props.passenger = !!this.props.passenger;
    if (!this.props.driver && !this.props.passenger) {
      this.props.driver = defaultDriverAndPassengerParameters.driver;
      this.props.seatsProposed =
        defaultDriverAndPassengerParameters.seatsProposed;
      this.props.passenger = defaultDriverAndPassengerParameters.passenger;
      this.props.seatsRequested =
        defaultDriverAndPassengerParameters.seatsRequested;
      return this;
    }
    if (!this.props.seatsProposed || this.props.seatsProposed <= 0)
      this.props.seatsProposed =
        defaultDriverAndPassengerParameters.seatsProposed;
    if (!this.props.seatsRequested || this.props.seatsRequested <= 0)
      this.props.seatsRequested =
        defaultDriverAndPassengerParameters.seatsRequested;
    return this;
  };

  private setMissingWaypointsPosition = (): AdEntity => {
    if (this.props.waypoints[0].position === undefined) {
      for (let i = 0; i < this.props.waypoints.length; i++) {
        this.props.waypoints[i].position = i;
      }
    }
    return this;
  };

  validate(): void {
    // entity business rules validation to protect it's invariant before saving entity to a database
  }
}

interface DefaultDriverAndPassengerParameters {
  driver: boolean;
  passenger: boolean;
  seatsProposed: number;
  seatsRequested: number;
}
